import java.io.BufferedWriter;
import java.util.ArrayList;
import java.util.LinkedList;

import aiinterface.AIInterface;
import aiinterface.CommandCenter;
import enumerate.Action;
import enumerate.State;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;
import struct.MotionData;
/**
 * Author Man-Je Kim, Donghyeon Lee
 */
public class JayBot_GM implements AIInterface {

   private Simulator simulator;
   private Key key;
   private CommandCenter commandCenter;
   private boolean playerNumber;
   private GameData gameData;
   private FrameData frameData;
   private FrameData simulatorAheadFrameData;
   private LinkedList<Action> myActions;
   private LinkedList<Action> oppActions;
   private CharacterData myCharacter;
   private CharacterData oppCharacter;
   private static final int FRAME_AHEAD = 14;
   private ArrayList<MotionData> myMotion;
   private ArrayList<MotionData> oppMotion;
   private Action[] actionAir;
   private Action[] actionGround;
   private Action[] myaction;
   private Action[] myactionZ;
   private Action[] myaction_100;
   private Action[] myaction_50;
   private Action[] myaction_85;
   private Action spSkill;
   private Node rootNode;
   private CharacterName charName;

  public static final boolean DEBUG_MODE = false;

   @Override
   public void close() {
   }

   @Override
   public void getInformation(FrameData frameData) {
      this.frameData = frameData;
      this.commandCenter.setFrameData(this.frameData, playerNumber);

      myCharacter = frameData.getCharacter(playerNumber);
      oppCharacter = frameData.getCharacter(!playerNumber);
   }

   @Override
   public int initialize(GameData gameData, boolean playerNumber) {
      this.playerNumber = playerNumber;
      this.gameData = gameData;

      this.key = new Key();
      this.frameData = new FrameData();
      this.commandCenter = new CommandCenter();

      this.myActions = new LinkedList<Action>();
      this.oppActions = new LinkedList<Action>();

      simulator = gameData.getSimulator();

      actionAir = new Action[] { Action.AIR_GUARD, Action.AIR_A, Action.AIR_B, Action.AIR_DA, Action.AIR_DB,
            Action.AIR_FA, Action.AIR_FB, Action.AIR_UA, Action.AIR_UB, Action.AIR_D_DF_FA, Action.AIR_D_DF_FB,
            Action.AIR_F_D_DFA, Action.AIR_F_D_DFB, Action.AIR_D_DB_BA, Action.AIR_D_DB_BB };
      actionGround = new Action[] { Action.STAND_D_DB_BA, Action.BACK_STEP, Action.FORWARD_WALK, Action.DASH,
            Action.JUMP, Action.FOR_JUMP, Action.BACK_JUMP, Action.STAND_GUARD, Action.CROUCH_GUARD, Action.THROW_A,
            Action.THROW_B, Action.STAND_A, Action.STAND_B, Action.CROUCH_A, Action.CROUCH_B, Action.STAND_FA,
            Action.STAND_FB, Action.CROUCH_FA, Action.CROUCH_FB, Action.STAND_D_DF_FA, Action.STAND_D_DF_FB,
            Action.STAND_F_D_DFA, Action.STAND_F_D_DFB, Action.STAND_D_DB_BB };
      spSkill = Action.STAND_D_DF_FC;

      myactionZ= new Action[] {Action.DASH	,Action.STAND_B	,Action.STAND_D_DB_BB		,Action.STAND_FA	, Action.FORWARD_WALK							 
		};
      myaction= new Action[] {
    		  Action.CROUCH_A,Action.FOR_JUMP,Action.STAND_D_DB_BB,Action.STAND_FB,Action.STAND_GUARD

      };
      myaction_50= new Action[] {
    		  Action.STAND_B,Action.STAND_D_DF_FB,Action.CROUCH_B,Action.STAND_D_DF_FA,Action.STAND_D_DB_BA

      };
      myaction_85= new Action[] {
    		  Action.BACK_STEP,Action.CROUCH_A,Action.STAND_D_DB_BB,Action.STAND_D_DF_FA,Action.STAND_GUARD

      };
      myaction_100= new Action[] {
    		  Action.CROUCH_FB,Action.STAND_A,Action.STAND_B,Action.STAND_D_DB_BA,Action.STAND_D_DF_FB

      };

      myMotion = gameData.getMotionData(this.playerNumber);
      oppMotion = gameData.getMotionData(!this.playerNumber);

      String tmpcharname = this.gameData.getCharacterName(this.playerNumber);
      if (tmpcharname.equals("ZEN"))
         charName = CharacterName.ZEN;
      else if (tmpcharname.equals("GARNET"))
         charName = CharacterName.GARNET;
      else if (tmpcharname.equals("LUD"))
         charName = CharacterName.LUD;
      else
         charName = CharacterName.OTHER;
      return 0;
   }

   @Override
   public Key input() {
      return key;
   }

   @Override
   public void processing() {
      if (canProcessing()) {
         if (commandCenter.getSkillFlag()) {
            key = commandCenter.getSkillKey();
         } else {
            key.empty();
            commandCenter.skillCancel();

            mctsPrepare(); 
            
          
            rootNode = new Node(simulatorAheadFrameData, null, myActions, oppActions, gameData, playerNumber,
                  commandCenter);
            rootNode.createNode();

            Action bestAction = rootNode.mcts();
            if (JayBot_GM.DEBUG_MODE) {
               rootNode.printNode(rootNode);
            }

            
            if(myCharacter.getCenterX()<=30||myCharacter.getCenterX()>=930)
            	commandCenter.commandCall("9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9");
            if(bestAction.name()=="STAND_GUARD")
            	commandCenter.commandCall("4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4");
            else if(bestAction.name()=="CROUCH_GUARD")
            	commandCenter.commandCall("1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1");
            else
            	commandCenter.commandCall(bestAction.name());

         }
      }
   }

  
   public boolean canProcessing() {
      return !frameData.getEmptyFlag() && frameData.getRemainingFramesNumber() > 0;
   }
   
   public void mctsPrepare() {
     simulatorAheadFrameData = simulator.simulate(frameData, playerNumber, null, null, FRAME_AHEAD);

      myCharacter = simulatorAheadFrameData.getCharacter(playerNumber);
      oppCharacter = simulatorAheadFrameData.getCharacter(!playerNumber);

      setMyAction();
      setOppAction();
   }

   public void setMyAction() {
      myActions.clear();

      int energy = myCharacter.getEnergy();
      if (charName.name() == "ZEN") {
			if (myCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionAir[i]);
						myActions.add(Action.CROUCH_GUARD);
		                myActions.add(Action.STAND_GUARD);
					}
				}
			} 	
			 else {
				
				if (Math.abs(
						myMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					myActions.add(spSkill);
				}
				for (int i = 0; i < myactionZ.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(myactionZ[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(myactionZ[i]);
					}
				}						
			}
      }
      else if (charName.name() == "LUD") {
			if (myCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionAir[i]);
					}
				}
			} else if (simulatorAheadFrameData.getDistanceX() < 140) {
				myActions.add(Action.AIR_DB);
				myActions.add(Action.AIR_B);
				myActions.add(Action.CROUCH_A);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.AIR_A);
				myActions.add(Action.CROUCH_FA);
				myActions.add(Action.STAND_B);
				if (myCharacter.getEnergy() >= 1)
					myActions.add(Action.STAND_A);
				if (myCharacter.getEnergy() >= 8)
					myActions.add(Action.STAND_FA);
				if (myCharacter.getEnergy() >= 10)
					myActions.add(Action.THROW_A);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.STAND_D_DB_BB);
			}
			else if (simulatorAheadFrameData.getDistanceX() < 190) {

				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.AIR_B);
				myActions.add(Action.CROUCH_A);
				myActions.add(Action.CROUCH_FA);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DB_BA);

				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DF_FA);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.STAND_D_DF_FB);
				if (myCharacter.getEnergy() >= 10)
					myActions.add(Action.AIR_UB);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.STAND_FB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_F_D_DFA);

			} else if (simulatorAheadFrameData.getDistanceX() < 240) {

				myActions.add(Action.CROUCH_B);
				myActions.add(Action.CROUCH_FA);
				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_A);
				myActions.add(Action.AIR_B);
				if (myCharacter.getEnergy() >= 8)
					myActions.add(Action.STAND_FA);
				if (myCharacter.getEnergy() >= 1)
					myActions.add(Action.STAND_A);
				if (myCharacter.getEnergy() >= 10)
					myActions.add(Action.THROW_A);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.CROUCH_FB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_F_D_DFA);
			} else {
				if (Math.abs(
						myMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					myActions.add(spSkill);
				}

				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionGround[i]);
					}
				}
			}
		}
      else if (charName.name() == "GARNET") {
         if (myCharacter.getState() == State.AIR) {
            for (int i = 0; i < actionAir.length; i++) {
               if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
                     .getAttackStartAddEnergy()) <= energy) {
            	   myActions.add(actionAir[i]);
               }
            }
         } else {
            if (Math.abs(
                  myMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
               myActions.add(spSkill);
            }

            if(simulatorAheadFrameData.getDistanceX()<=105)
            {
               myActions.add(Action.CROUCH_GUARD);
               myActions.add(Action.STAND_GUARD);
               for (int i = 0; i < myaction_50.length; i++) {
                  if (Math.abs(myMotion.get(Action.valueOf(myaction_50[i].name()).ordinal())
                        .getAttackStartAddEnergy()) <= energy) {
                     myActions.add(myaction_50[i]);
                  }
               }
            }
            else if(simulatorAheadFrameData.getDistanceX()<=120)
            {
            	myActions.add(Action.CROUCH_GUARD);
                myActions.add(Action.STAND_GUARD);
               
               for (int i = 0; i < myaction_85.length; i++) {
                  if (Math.abs(myMotion.get(Action.valueOf(myaction_85[i].name()).ordinal())
                        .getAttackStartAddEnergy()) <= energy) {
                     myActions.add(myaction_85[i]);
                  }
               }
            }
            else if(simulatorAheadFrameData.getDistanceX()<=200)
            {
            	myActions.add(Action.CROUCH_GUARD);
                myActions.add(Action.STAND_GUARD);
               for (int i = 0; i < myaction_100.length; i++) {
                  if (Math.abs(myMotion.get(Action.valueOf(myaction_100[i].name()).ordinal())
                        .getAttackStartAddEnergy()) <= energy) {
                     myActions.add(myaction_100[i]);
                  }
               }
            }
           else {
               for (int i = 0; i < myaction.length; i++) {
                  if (Math.abs(myMotion.get(Action.valueOf(myaction[i].name()).ordinal())
                        .getAttackStartAddEnergy()) <= energy) {
                     myActions.add(myaction[i]);
                  }
               }
            }
         }
      }
   }

   public void setOppAction() {
      oppActions.clear();
      int energy = oppCharacter.getEnergy();
      if (oppCharacter.getState() == State.AIR) {
         for (int i = 0; i < actionAir.length; i++) {
            if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
                  .getAttackStartAddEnergy()) <= energy) {
               oppActions.add(actionAir[i]);
            }
         }
      } else {
         if (Math.abs(oppMotion.get(Action.valueOf(spSkill.name()).ordinal())
               .getAttackStartAddEnergy()) <= energy) {
            oppActions.add(spSkill);
         }

         for (int i = 0; i < actionGround.length; i++) {
            if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
                  .getAttackStartAddEnergy()) <= energy) {
               oppActions.add(actionGround[i]);
            }
         }
      }
   }
   public boolean EqulCheck(Action a, Action b)
   {
        if (a==b)
           return true;
        else
           return false;
   }

   @Override
   public void roundEnd(int p1Hp, int p2Hp, int frames) {
      commandCenter.skillCancel();
      key.empty();
   }

     public enum CharacterName {
         ZEN,GARNET,LUD,OTHER;
      }
}
